package com.vaadin.spring.roo.addon;

import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.Validate;
import org.springframework.roo.addon.jpa.activerecord.JpaActiveRecordMetadata;
import org.springframework.roo.classpath.PhysicalTypeCategory;
import org.springframework.roo.classpath.PhysicalTypeDetails;
import org.springframework.roo.classpath.PhysicalTypeIdentifier;
import org.springframework.roo.classpath.PhysicalTypeIdentifierNamingUtils;
import org.springframework.roo.classpath.PhysicalTypeMetadata;
import org.springframework.roo.classpath.customdata.CustomDataKeys;
import org.springframework.roo.classpath.details.BeanInfoUtils;
import org.springframework.roo.classpath.details.ClassOrInterfaceTypeDetails;
import org.springframework.roo.classpath.details.FieldMetadata;
import org.springframework.roo.classpath.details.MemberFindingUtils;
import org.springframework.roo.classpath.details.MethodMetadata;
import org.springframework.roo.classpath.scanner.MemberDetails;
import org.springframework.roo.classpath.scanner.MemberDetailsScanner;
import org.springframework.roo.metadata.MetadataIdentificationUtils;
import org.springframework.roo.metadata.MetadataService;
import org.springframework.roo.model.JavaSymbolName;
import org.springframework.roo.model.JavaType;
import org.springframework.roo.process.manager.FileManager;
import org.springframework.roo.project.LogicalPath;
import org.springframework.roo.project.Path;

/**
 * Utility methods used by the Vaadin Roo addon.
 */
public abstract class VaadinRooUtils {

	public static void installFromTemplateIfNeeded(FileManager fileManager,
			String destinationFile, String templateBaseName,
			Map<String, String> substitutions) {
		if (!fileManager.exists(destinationFile)) {
			InputStream templateInputStream = org.springframework.roo.support.util.FileUtils
					.getInputStream(VaadinRooUtils.class, templateBaseName
							+ "-template");
			try {
				// Read template and insert the user's package
				String input = IOUtils.toString(templateInputStream);
				for (Map.Entry<String, String> entry : substitutions.entrySet()) {
					input = input.replace(entry.getKey(), entry.getValue());
				}

				// Output the file for the user
				fileManager.createOrUpdateTextFileIfRequired(destinationFile,
						input, true);
			} catch (IOException ioe) {
				throw new IllegalStateException("Unable to create '"
						+ destinationFile + "'", ioe);
			} finally {
				IOUtils.closeQuietly(templateInputStream);
			}
		}
	}

	/**
	 * Checks if a field exists (directly) in a type.
	 * 
	 * @param fieldName
	 * @param governorTypeDetails
	 * @return field found or null if none
	 */
	public static FieldMetadata fieldExists(JavaSymbolName fieldName,
			ClassOrInterfaceTypeDetails governorTypeDetails) {
		// We have no access to method parameter information, so we scan by name
		// alone and treat any match as authoritative
		// We do not scan the superclass, as the caller is expected to know
		// we'll only scan the current class
		for (FieldMetadata field : governorTypeDetails.getDeclaredFields()) {
			if (field.getFieldName().equals(fieldName)) {
				// Found a field with the expected name
				return field;
			}
		}
		return null;
	}

	/**
	 * Checks if a method exists (directly) in a type. Used to avoid duplicate
	 * method generation when creating ITD.
	 * 
	 * @param methodName
	 * @param governorTypeDetails
	 * @return method found or null if none
	 */
	public static MethodMetadata methodExists(JavaSymbolName methodName,
			ClassOrInterfaceTypeDetails governorTypeDetails) {
		// We have no access to method parameter information, so we scan by name
		// alone and treat any match as authoritative
		// We do not scan the superclass, as the caller is expected to know
		// we'll only scan the current class
		for (MethodMetadata method : governorTypeDetails.getDeclaredMethods()) {
			if (method.getMethodName().equals(methodName)) {
				// Found a method of the expected name; we won't check method
				// parameters though
				return method;
			}
		}
		return null;
	}

	/**
	 * @deprecated use Roo's getMemberDetails instead
	 * 
	 *             Gets the member details for a class.
	 * 
	 * @param type
	 * @param metadataService
	 * @param memberDetailsScanner
	 * @param requestingClass
	 *            class name requesting the accessors, usually
	 *            this.getClass().getName()
	 * @return
	 */
	@Deprecated
	public static MemberDetails getMemberDetails(JavaType type,
			MetadataService metadataService,
			MemberDetailsScanner memberDetailsScanner, String requestingClass) {
		// TODO use logical path for type?
		LogicalPath path = LogicalPath.getInstance(Path.SRC_MAIN_JAVA, "");
		PhysicalTypeMetadata physicalTypeMetadata = (PhysicalTypeMetadata) metadataService
				.get(PhysicalTypeIdentifier.createIdentifier(type, path));
		Validate.notNull(physicalTypeMetadata,
				"The type " + type.getFullyQualifiedTypeName()
						+ " does not exist or can not be found.");
		PhysicalTypeDetails ptd = physicalTypeMetadata
				.getMemberHoldingTypeDetails();
		Validate.isInstanceOf(ClassOrInterfaceTypeDetails.class, ptd);
		ClassOrInterfaceTypeDetails classOrInterfaceTypeDetails = (ClassOrInterfaceTypeDetails) ptd;

		return memberDetailsScanner.getMemberDetails(requestingClass,
				classOrInterfaceTypeDetails);
	}

	/**
	 * Returns the properties of a type and the corresponding accessors (public
	 * parameterless get/is methods), excluding accessors for version and id
	 * properties of an entity.
	 * 
	 * @param type
	 * @param memberDetails
	 * @param withMutatorOnly
	 *            only return accessors that have a mutator
	 * @return map from property name to corresponding getter method metadata
	 */
	public static Map<JavaSymbolName, MethodMetadata> getAccessors(
			JavaType type, MemberDetails memberDetails, boolean withMutatorOnly) {
		// TODO there should be a better API for getting these

		// normal properties of the type, excluding ID and version properties
		Map<JavaSymbolName, MethodMetadata> properties = new LinkedHashMap<JavaSymbolName, MethodMetadata>();
		for (MethodMetadata method : MemberFindingUtils
				.getMethods(memberDetails)) {

			if (BeanInfoUtils.isAccessorMethod(method)
					&& !method
							.getCustomData()
							.keySet()
							.contains(CustomDataKeys.IDENTIFIER_ACCESSOR_METHOD)
					&& !method.getCustomData().keySet()
							.contains(CustomDataKeys.VERSION_ACCESSOR_METHOD)) {
				JavaSymbolName propertyName = BeanInfoUtils
						.getPropertyNameForJavaBeanMethod(method);
				if (null != propertyName) {
					final FieldMetadata field = BeanInfoUtils
							.getFieldForJavaBeanMethod(memberDetails, method);
					if (field != null
							&& (!withMutatorOnly || BeanInfoUtils
									.hasAccessorAndMutator(field, memberDetails))) {
						properties.put(propertyName, method);
					}
				}
			}
		}
		return properties;
	}

	public static Map<JavaSymbolName, JavaType> getSpecialDomainTypes(
			MetadataService metadataService, JavaType javaType,
			MemberDetails memberDetails, boolean includeEmbedded) {
		Map<JavaSymbolName, JavaType> specialTypes = new LinkedHashMap<JavaSymbolName, JavaType>();

		if (!isEntityTypeInProject(metadataService, javaType)) {
			// unable to get entity metadata so it is not a Entity in our
			// project anyway.
			return specialTypes;
		}

		Map<JavaSymbolName, MethodMetadata> accessors = VaadinRooUtils
				.getAccessors(javaType, memberDetails, true);

		for (JavaSymbolName propertyName : accessors.keySet()) {
			JavaType type = accessors.get(propertyName).getReturnType();

			if (type.isCommonCollectionType()) {
				for (JavaType genericType : type.getParameters()) {
					if (isDomainTypeInProject(metadataService, genericType)) {
						specialTypes.put(propertyName, genericType);
					}
				}
			} else {
				FieldMetadata fieldMetadata = BeanInfoUtils
						.getFieldForPropertyName(memberDetails, propertyName);

				if (isDomainTypeInProject(metadataService, type)
						&& (includeEmbedded || !isEmbeddedFieldType(fieldMetadata))) {
					specialTypes.put(propertyName, type);
				}
			}
		}
		return specialTypes;
	}

	public static boolean isEnumType(MetadataService metadataService,
			JavaType type) {
		// TODO use logical path for javaType?
		LogicalPath path = LogicalPath.getInstance(Path.SRC_MAIN_JAVA, "");
		PhysicalTypeMetadata physicalTypeMetadata = (PhysicalTypeMetadata) metadataService
				.get(PhysicalTypeIdentifierNamingUtils.createIdentifier(
						PhysicalTypeIdentifier.class.getName(), type, path));
		if (physicalTypeMetadata != null) {
			PhysicalTypeDetails details = physicalTypeMetadata
					.getMemberHoldingTypeDetails();
			if (details != null) {
				if (details.getPhysicalTypeCategory().equals(
						PhysicalTypeCategory.ENUMERATION)) {
					return true;
				}
			}
		}
		return false;
	}

	public static boolean isDomainTypeInProject(
			MetadataService metadataService, JavaType javaType) {
		// we are only interested if the type is part of our application
		// TODO use logical path for javaType?
		LogicalPath path = LogicalPath.getInstance(Path.SRC_MAIN_JAVA, "");
		if (metadataService.get(PhysicalTypeIdentifier.createIdentifier(
				javaType, path)) != null) {
			return true;
		}
		return false;
	}

	public static boolean isEntityTypeInProject(
			MetadataService metadataService, JavaType javaType) {
		// we are only interested if the type is part of our application
		// TODO use logical path for javaType?
		LogicalPath path = LogicalPath.getInstance(Path.SRC_MAIN_JAVA, "");
		if (metadataService.get(JpaActiveRecordMetadata.createIdentifier(
				javaType, path)) != null) {
			return true;
		}
		return false;
	}

	public static boolean isEmbeddedFieldType(FieldMetadata field) {
		return MemberFindingUtils.getAnnotationOfType(field.getAnnotations(),
				new JavaType("javax.persistence.Embedded")) != null;
	}

	public static boolean isNotificationForJavaType(String mid) {
		return MetadataIdentificationUtils.getMetadataClass(mid).equals(
				MetadataIdentificationUtils
						.getMetadataClass(PhysicalTypeIdentifier
								.getMetadataIdentiferType()));
	}

}
