package com.vaadin.spring.roo.addon.entityview;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.logging.Logger;

import org.apache.felix.scr.annotations.Component;
import org.apache.felix.scr.annotations.Reference;
import org.apache.felix.scr.annotations.Service;
import org.osgi.service.component.ComponentContext;
import org.springframework.roo.addon.jpa.activerecord.JpaActiveRecordMetadata;
import org.springframework.roo.addon.web.mvc.controller.details.JavaTypeMetadataDetails;
import org.springframework.roo.addon.web.mvc.controller.details.WebMetadataService;
import org.springframework.roo.classpath.PhysicalTypeIdentifier;
import org.springframework.roo.classpath.PhysicalTypeMetadata;
import org.springframework.roo.classpath.customdata.CustomDataKeys;
import org.springframework.roo.classpath.customdata.tagkeys.MethodMetadataCustomDataKey;
import org.springframework.roo.classpath.details.ClassOrInterfaceTypeDetails;
import org.springframework.roo.classpath.details.ItdTypeDetails;
import org.springframework.roo.classpath.details.MemberFindingUtils;
import org.springframework.roo.classpath.details.MemberHoldingTypeDetails;
import org.springframework.roo.classpath.itd.AbstractMemberDiscoveringItdMetadataProvider;
import org.springframework.roo.classpath.itd.ItdTriggerBasedMetadataProvider;
import org.springframework.roo.classpath.itd.ItdTypeDetailsProvidingMetadataItem;
import org.springframework.roo.classpath.layers.MemberTypeAdditions;
import org.springframework.roo.classpath.scanner.MemberDetails;
import org.springframework.roo.model.JavaType;
import org.springframework.roo.project.LogicalPath;
import org.springframework.roo.project.Path;
import org.springframework.roo.support.util.CollectionUtils;

import com.vaadin.spring.roo.addon.VaadinRooUtils;
import com.vaadin.spring.roo.addon.abstractentityview.VaadinAbstractEntityViewMetadata;
import com.vaadin.spring.roo.addon.annotations.RooVaadinAbstractEntityView;
import com.vaadin.spring.roo.addon.annotations.RooVaadinEntityView;

/**
 * Provides {@link VaadinEntityViewMetadata}.
 */
@Component(immediate = true)
@Service
public class VaadinEntityViewMetadataProvider extends
		AbstractMemberDiscoveringItdMetadataProvider implements
		ItdTriggerBasedMetadataProvider {
	// XXX see {@link WebScaffoldMetadataProviderImpl} or {@link
	// JsfManagedBeanMetadataImpl} .
	// public class VaadinEntityViewMetadataProvider extends
	// AbstractItdMetadataProvider {

	@Reference
	private WebMetadataService webMetadataService;

	private final Map<JavaType, String> entityToViewMidMap = new LinkedHashMap<JavaType, String>();
	private final Map<String, JavaType> viewMidToEntityMap = new LinkedHashMap<String, JavaType>();

	private static Logger logger = Logger
			.getLogger(VaadinEntityViewMetadataProvider.class.getName());

	protected void activate(ComponentContext context) {
		metadataDependencyRegistry.addNotificationListener(this);
		metadataDependencyRegistry.registerDependency(
				PhysicalTypeIdentifier.getMetadataIdentiferType(),
				getProvidesType());
		addMetadataTrigger(new JavaType(RooVaadinEntityView.class.getName()));
	}

	protected void deactivate(ComponentContext context) {
		metadataDependencyRegistry.removeNotificationListener(this);
		metadataDependencyRegistry.deregisterDependency(
				PhysicalTypeIdentifier.getMetadataIdentiferType(),
				getProvidesType());
		removeMetadataTrigger(new JavaType(RooVaadinEntityView.class.getName()));
	}

	@Override
	protected String createLocalIdentifier(JavaType javaType, LogicalPath path) {
		return VaadinEntityViewMetadata.createIdentifier(javaType, path);
	}

	@Override
	protected String getGovernorPhysicalTypeIdentifier(
			String metadataIdentificationString) {
		JavaType javaType = VaadinEntityViewMetadata
				.getJavaType(metadataIdentificationString);
		LogicalPath path = VaadinEntityViewMetadata
				.getPath(metadataIdentificationString);
		return PhysicalTypeIdentifier.createIdentifier(javaType, path);
	}

	public String getItdUniquenessFilenameSuffix() {
		return "VaadinEntityView";
	}

	/**
	 * From {@link AbstractMemberDiscoveringItdMetadataProvider}: Allows a
	 * subclass to assess a recently updated {@link ItdTypeDetails} and decide
	 * whether they are interested in a metadata request being made in response.
	 * Subclasses will generally iterate over the passed ITD details and search
	 * for members of interest. If any members of interest are located,
	 * subclasses will return an instance-specific metadata identification
	 * string (MID) consistent with the subclass' {@link #getProvidesType()} The
	 * requested MID will be cleared from the cache and formally requested. This
	 * process allows subclasses to effectively discover new ITD members that
	 * appear over time without needing to process every request themselves.
	 * 
	 * @param itdTypeDetails
	 *            a valid {@link ItdTypeDetails} from which member information
	 *            is available (never null)
	 * @return null if the subclass is not interested in the type, or a MID if
	 *         it is
	 */
	@Override
	protected String getLocalMidToRequest(ItdTypeDetails itdTypeDetails) {
		// XXX implementation from {@link
		// JsfManagedBeanMetadataProviderImpl#getLocalMidToRequest}
		// Determine the governor for this ITD, and whether any metadata is even
		// hoping to hear about changes to that JavaType and its ITDs
		final JavaType governor = itdTypeDetails.getName();
		final String localMid = entityToViewMidMap.get(governor);
		if (localMid != null) {
			return localMid;
		}

		final MemberHoldingTypeDetails memberHoldingTypeDetails = typeLocationService
				.getTypeDetails(governor);
		if (memberHoldingTypeDetails != null) {
			for (final JavaType type : memberHoldingTypeDetails
					.getLayerEntities()) {
				final String localMidType = entityToViewMidMap.get(type);
				if (localMidType != null) {
					return localMidType;
				}
			}
		}
		return null;
	}

	@Override
	protected ItdTypeDetailsProvidingMetadataItem getMetadata(
			String metadataIdentificationString, JavaType aspectName,
			PhysicalTypeMetadata governorPhysicalTypeMetadata,
			String itdFilename) {
		// We need to parse the annotation, which we expect to be present
		VaadinEntityViewAnnotationValues annotationValues = new VaadinEntityViewAnnotationValues(
				governorPhysicalTypeMetadata);
		final JavaType formBackingType = annotationValues.formBackingObject;
		if (!annotationValues.isAnnotationFound() || formBackingType == null) {
			return null;
		}

		final MemberDetails formBackingObjectMemberDetails = getMemberDetails(formBackingType);
		if (formBackingObjectMemberDetails == null) {
			return null;
		}

		final MemberHoldingTypeDetails formBackingMemberHoldingTypeDetails = MemberFindingUtils
				.getMostConcreteMemberHoldingTypeDetailsWithTag(
						formBackingObjectMemberDetails,
						CustomDataKeys.PERSISTENT_TYPE);
		if (formBackingMemberHoldingTypeDetails == null) {
			return null;
		}

		final Map<MethodMetadataCustomDataKey, MemberTypeAdditions> crudAdditions = webMetadataService
				.getCrudAdditions(formBackingType, metadataIdentificationString);
		if (CollectionUtils.isEmpty(crudAdditions)) {
			return null;
		}

		// Lookup the form backing object's metadata
		// TODO use logical path for formBackingType?
		LogicalPath path = LogicalPath.getInstance(Path.SRC_MAIN_JAVA, null);

		String physicalTypeMetadataKey = PhysicalTypeIdentifier
				.createIdentifier(formBackingType, path);
		String entityMetadataKey = JpaActiveRecordMetadata.createIdentifier(
				formBackingType, path);

		// We need to lookup the metadata we depend on
		JpaActiveRecordMetadata entityMetadata = (JpaActiveRecordMetadata) metadataService
				.get(entityMetadataKey);

		// We need to abort if we couldn't find dependent metadata
		if (entityMetadata == null || !entityMetadata.isValid()) {
			return null;
		}

		// We need to be informed if our dependent metadata changes
		metadataDependencyRegistry.registerDependency(physicalTypeMetadataKey,
				metadataIdentificationString);
		metadataDependencyRegistry.registerDependency(entityMetadataKey,
				metadataIdentificationString);
		metadataDependencyRegistry
				.registerDependency(JpaActiveRecordMetadata.createIdentifier(
						formBackingType, path), metadataIdentificationString);

		// We also need to be informed if a referenced type is changed
		Collection<JavaType> specialDomainTypes = VaadinRooUtils
				.getSpecialDomainTypes(metadataService, formBackingType,
						formBackingObjectMemberDetails, true).values();
		// unique types only
		for (JavaType type : new LinkedHashSet<JavaType>(specialDomainTypes)) {
			metadataDependencyRegistry.registerDependency(
					PhysicalTypeIdentifier.createIdentifier(type, path),
					metadataIdentificationString);
			metadataDependencyRegistry.registerDependency(
					JpaActiveRecordMetadata.createIdentifier(type, path),
					metadataIdentificationString);
		}

		// find superclasses up to one that has the annotation
		// @RooVaadinAbstractEntityView

		// this is a list of physical type metadata IDs for the superclasses of
		// the governor type and ending with that of the class with the
		// @RooVaadinAbstractEntityView annotation, or null if no such
		// superclass is found
		List<String> superclassPhysicalTypeMetadataIds = getSuperclassPhysicalTypeMetadataIds(metadataIdentificationString);
		if (superclassPhysicalTypeMetadataIds == null) {
			return null;
		}

		// register metadata dependencies on all the superclasses to monitor
		// changes to them
		for (String mid : superclassPhysicalTypeMetadataIds) {
			metadataDependencyRegistry.registerDependency(mid,
					metadataIdentificationString);
		}

		// get VaadinAbstractEntityViewMetadata for the "last" superclass
		String abstractViewPhysicalTypeId = superclassPhysicalTypeMetadataIds
				.get(superclassPhysicalTypeMetadataIds.size() - 1);
		String abstractViewId = VaadinAbstractEntityViewMetadata
				.createIdentifier(PhysicalTypeIdentifier
						.getJavaType(abstractViewPhysicalTypeId),
						PhysicalTypeIdentifier
								.getPath(abstractViewPhysicalTypeId));
		VaadinAbstractEntityViewMetadata abstractViewMetadata = (VaadinAbstractEntityViewMetadata) metadataService
				.get(abstractViewId);
		if (abstractViewMetadata == null) {
			return null;
		}

		// Remember that this entity JavaType matches up with this metadata
		// identification string
		// Start by clearing any previous association
		final JavaType oldEntity = viewMidToEntityMap
				.get(metadataIdentificationString);
		if (oldEntity != null) {
			entityToViewMidMap.remove(oldEntity);
		}
		entityToViewMidMap.put(formBackingType, metadataIdentificationString);
		viewMidToEntityMap.put(metadataIdentificationString, formBackingType);

		final SortedMap<JavaType, JavaTypeMetadataDetails> relatedApplicationTypeMetadata = webMetadataService
				.getRelatedApplicationTypeMetadata(formBackingType,
						formBackingObjectMemberDetails,
						metadataIdentificationString);
		final JavaTypeMetadataDetails javaTypeMetadataHolder = relatedApplicationTypeMetadata
				.get(formBackingType);

		// We do not need to monitor the parent, as any changes to the java type
		// associated with the parent will trickle down to
		// the governing java type
		return new VaadinEntityViewMetadata(metadataIdentificationString,
				aspectName, governorPhysicalTypeMetadata, metadataService,
				memberDetailsScanner, webMetadataService, typeLocationService,
				annotationValues,
				new VaadinEntityMetadataDetails(formBackingType,
						entityMetadata, javaTypeMetadataHolder),
				abstractViewMetadata);
	}

	private List<String> getSuperclassPhysicalTypeMetadataIds(
			String metadataIdentificationString) {
		List<String> superclassPhysicalTypeMetadataIds = new ArrayList<String>();

		PhysicalTypeMetadata ptm = (PhysicalTypeMetadata) metadataService
				.get(getGovernorPhysicalTypeIdentifier(metadataIdentificationString));
		if (ptm == null
				|| !(ptm.getMemberHoldingTypeDetails() instanceof ClassOrInterfaceTypeDetails)) {
			return null;
		}

		JavaType abstractEntityViewAnnotationType = new JavaType(
				RooVaadinAbstractEntityView.class.getName());

		ClassOrInterfaceTypeDetails cid = ptm.getMemberHoldingTypeDetails();
		ClassOrInterfaceTypeDetails superclass = cid;
		while (true) {
			superclass = superclass.getSuperclass();
			if (superclass == null) {
				return null;
			}
			String superclassPhysicalMetadataId = superclass
					.getDeclaredByMetadataId();
			superclassPhysicalTypeMetadataIds.add(superclassPhysicalMetadataId);
			// break if superclass has @RooVaadinAbstractEntityView
			if (superclass.getAnnotation(abstractEntityViewAnnotationType) != null) {
				break;
			}
		}
		return superclassPhysicalTypeMetadataIds;
	}

	public String getProvidesType() {
		return VaadinEntityViewMetadata.getMetadataIdentiferType();
	}

}
